from os import listdir
from re import search, match
from subprocess import Popen, PIPE


def available_cpu_count():
    """
    Determine the number of available virtual or physical CPUs on the system,
    running the scripts, i.e. user/real as output by time(1) when called with
    an optimally scaling userspace-only program.
    :return: int, number of available virtual or physical CPUs
    """

    # cpuset, preferred if available
    # Note that: cpuset may restrict the number of *available* processors
    try:
        with open('/proc/self/status') as f:
            m = search(r'(?m)^Cpus_allowed:\s*(.*)$', f.read())
        if m:
            res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
            if res > 0:
                return res
    except IOError:
        pass

    # Python 2.6+, covers most use cases where cpuset is not available
    try:
        import multiprocessing
        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    # https://github.com/giampaolo/psutil preferred fallback, if installed
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from psutil import cpu_count
        return cpu_count()
    except ImportError:
        try:
            # noinspection PyPackageRequirements,PyUnresolvedReferences
            from psutil import NUM_CPUS
            return NUM_CPUS
        except ImportError:
            pass

    # POSIX
    try:
        from os import sysconf
        res = int(sysconf('SC_NPROCESSORS_ONLN'))

        if res > 0:
            return res
    except (ImportError, ValueError):
        pass

    # Windows
    try:
        from os import environ
        res = int(environ['NUMBER_OF_PROCESSORS'])

        if res > 0:
            return res
    except (ImportError, KeyError, ValueError):
        pass

    # jython
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        from java.lang import Runtime

        runtime = Runtime.getRuntime()
        res = runtime.availableProcessors()

        if res > 0:
            return res
    except ImportError:
        pass

    # BSD
    try:
        sysctl = Popen(['sysctl', '-n', 'hw.ncpu'],
                       stdout=PIPE)
        sc_stdout = sysctl.communicate()[0]
        res = int(sc_stdout)

        if res > 0:
            return res
    except (OSError, ValueError):
        pass

    # Linux
    try:
        with open('/proc/cpuinfo') as f:
            res = f.read().count('processor\t:')

        if res > 0:
            return res
    except IOError:
        pass

    # Solaris
    try:
        pseudo_devices = listdir('/devices/pseudo/')
        res = 0
        for pd in pseudo_devices:
            if match(r'^cpuid@[0-9]+$', pd):
                res += 1

        if res > 0:
            return res
    except OSError:
        pass

    # Other UNIXes (heuristic)
    try:
        try:
            with open('/var/run/dmesg.boot') as f:
                dmesg = f.read()
        except IOError:
            dmesg_process = Popen(['dmesg'], stdout=PIPE)
            dmesg = dmesg_process.communicate()[0]

        res = 0
        while '\ncpu' + str(res) + ':' in dmesg:
            res += 1

        if res > 0:
            return res
    except OSError:
        pass

    raise Exception('Can not determine number of CPUs on this system')



from ._version import __version__
from ._py7za import Py7za, arg_split
from ._nice_size import nice_size, size_to_int
from ._asyncio_pool import AsyncIOPool
from ._cpu_count import available_cpu_count



from shlex import split
from typing import Union, List, Callable
from pathlib import Path
from asyncio import create_subprocess_exec, run
from asyncio.subprocess import PIPE
from shutil import which
from os import name as os_name
from datetime import datetime


def arg_split(args, platform=os_name):
    """
    Like calling shlex.split, but sets `posix=` according to platform
    and unquotes previously quoted arguments on Windows
    :param args: a command line string consisting of a command with arguments, e.g. 'dir "C:/Program Files"'
    :param platform: a value like os.name would return, e.g. 'nt'
    :return: a list of arguments like shlex.split(args) would have returned
    """
    return [a[1:-1].replace('""', '"') if a[0] == a[-1] == '"' else a
            for a in (split(args, posix=False) if platform == 'nt' else split(args))]


class Py7za:
    """
    Wrapper class for running 7za.exe.

    Attributes
    ----------
    executable_7za: str (class)
        full file path to 7za.exe, executable from calling relative dir (or just pre-installed '7za' on non-Windows)
    progress: int
        7za operation progress
    files: List[str]
        files that were processed in the operation
    done: bool
        whether operation has completed
    errors: bytes
        stderr of operation, once it completes (or fails)
    """
    executable_7za = str(Path(__file__).parent / 'bin/7za.exe') if os_name == 'nt' else '7za'

    def __init__(self, arguments: Union[str, List[str]], on_start: Callable = None, working_dir: str = '.'):
        """
        Creates an (awaitable) object ready to run 7za with given arguments
        :param arguments: arguments to pass to 7za, after processing (always pass progress and output>1, disable log)
        :param on_start: callback to be called just before starting a 7za subprocess is started
        :param working_dir: working directory for 7za
        """
        if which(self.executable_7za) is None:
            raise FileNotFoundError(f'7za executable "{self.executable_7za}" not found.')

        if isinstance(arguments, str):
            # replacing backslashes with forward slashes, to allow split to interpret correctly; this works as long
            # as the Windows version supports it, but as this script only work on Python 3.8>, no issue
            arguments = arg_split(arguments)

        # ignore output arguments passed, always pass progress and output to 1, disable log
        self.arguments = [a for a in arguments if a[:3] not in ['-bs', '-bb']] + ['-bsp1', '-bso1', '-bb']

        self.progress = 0
        self.files = []

        self.done = False
        self.errors = None
        self.return_code = None

        self.working_dir = working_dir

        self.on_start = on_start

    def __await__(self):
        return self.arun().__await__()

    def _parse_stdout(self, line):
        if line:
            line = line.decode()
            if line[0] in '+U':
                self.files.append((line[0], line[2:]))
            if len(line) >= 4 and line[3] == '%':
                self.progress = int(line[:3].strip())

    async def arun(self) -> 'Py7za':
        """
        Run 7za asynchronously, updating .progress and .files during the run and .done and errors when it completes
        :return: self (with updated attributes, like .return_code and .errors)
        """
        self.progress = 0
        self.files = []

        self.done = False
        self.errors = None

        if self.on_start is not None:
            self.on_start(self)

        proc = await create_subprocess_exec(
            self.executable_7za, *self.arguments, stdout=PIPE, stderr=PIPE, cwd=self.working_dir)

        line = b''
        while True:
            b = await proc.stdout.read(1)
            if b == b'\r':
                self._parse_stdout(line)
                line = b''
            else:
                line += b
            # stdout read will return b'' when the process has ended
            if b == b'':
                self.errors = await proc.stderr.read()
                await proc.wait()
                self.done = True
                if proc.returncode == 0:
                    self.progress = 100
                self.return_code = proc.returncode
                return self

    def run(self) -> int:
        """
        Run and await arun()
        :return: return code of process
        """
        return run(self.arun())

    @classmethod
    async def list_archive(cls, archive: str, meta_data_only: bool = False) -> tuple:
        """
        Opens and lists contents of archive, returning either just metadata or metadata and a file list with details
        :param archive: name of the archive to list
        :param meta_data_only: whether to only return metadata, or parse and include archive contents details
        :return: tuple of size, compressed size, number of files, number of directories, and list of files
                 list of files is empty or a list of tuples of datetime, attributes, size, compressed size, and name
        """
        proc = await create_subprocess_exec(
            cls.executable_7za, 'l', archive, stdout=PIPE)
        line = '\n'
        listing = False
        files = []
        while line:
            line = await proc.stdout.readline()
            if line.startswith(b'----------'):
                line = await proc.stdout.readline()
                if listing:
                    break
                listing = not listing
            if not meta_data_only and listing:
                t = line.decode().split()
                if len(t) == 6:
                    d, t, a, s, c, n = t
                else:
                    d, t, a, s, n = t
                    c = 0
                files.append((datetime(*map(int, d.split('-')), *map(int, t.split(':'))), a, int(s), int(c), n))
        await proc.wait()
        metadata = line.decode().split()
        if not metadata:
            s, c, f, d = 0, 0, 0, 0
        elif len(metadata) == 8:
            __, __, s, c, f, __, d, __ = metadata
        else:
            __, __, s, c, f, __ = metadata
            d = 0
        return int(s), int(c), int(f), int(d), files



SI_UNITS = ['kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']
L_SI_UNITS = [unit.lower() for unit in SI_UNITS]
BIN_UNITS = ['KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
L_BIN_UNITS = [unit.lower() for unit in BIN_UNITS]


def nice_size(size: int, si: bool = False, decimal_precision: int = 1) -> str:
    """
    Returns a string representation of a file size in SI (KiB, MiB, etc.)
    or binary units (KB, MB, etc.)
    :param size: a size in single bytes
    :param si: whether to use SI units (or binary units, the default)
    :param decimal_precision: the number of decimals to show in rounded
        representations
    :return: a string representation of size
    """
    threshold = 1000 if si else 1024

    if abs(size) < threshold:
        return f'{size} B'

    units = SI_UNITS if si else BIN_UNITS
    u = -1
    r = 10 ** decimal_precision

    while True:
        size /= threshold
        u += 1
        if round(abs(size) * r) / r < threshold or u == len(units) - 1:
            break

    # noinspection PyStringFormat
    return (f'%.{decimal_precision}f ' % size) + units[u]


def size_to_int(s: str, si: bool = False) -> int:
    """
    Returns an integer value for a string as it would have been generated by
    nice_size (e.g. '900 B', '4.5 GiB', etc.)
    :param s: a string representing a size in SI or binary units
        (e.g. '900 B', '4.5 GiB', etc.)
    :param si: whether to use SI units (or binary units, the default)
    :return: a round integer value matching the nice size string
    """
    t = s.lower().strip().split()
    if len(t) != 2:
        raise ValueError(f'cannot convert {s} to number of bytes')
    elif t[1].lower() == 'b':
        return int(t[0])
    if si:
        return int(1000 ** (L_SI_UNITS.index(t[1]) + 1) * float(t[0]))
    else:
        return int(1024 ** (L_BIN_UNITS.index(t[1]) + 1) * float(t[0]))



import unittest
from asyncio import sleep
from random import random
from py7za.py7za_box import AsyncIOPool


class AwaitableCounter:
    number_running = 0
    completed = 0

    def __init__(self, n, t=.01, rnd=True):
        self.rnd = rnd
        self.t = t
        self.n = n
        self.running = False

    def __await__(self):
        return self.arun().__await__()

    def close(self):
        if self.running:
            self.number_running -= 1
            self.running = False

    async def arun(self) -> 'AwaitableCounter':
        self.running = True
        self.number_running += 1
        while self.n > 0:
            self.n -= 1
            await sleep(self.t if not self.rnd else self.t * 2 * random())
        self.completed += 1
        self.number_running -= 1
        return self


class TestPy7zaBox(unittest.IsolatedAsyncioTestCase):
    async def test_aiop(self):
        n = 4
        aiop = AsyncIOPool(n)
        tasks = [AwaitableCounter(10) for __ in range(10)]
        done = []
        async for t in aiop.arun_many(tasks):
            done.append(t)
            self.assertLess(AwaitableCounter.number_running, n, f'never more than pool size running')
        self.assertEqual(len(done), 10, 'all tasks completed')

    async def test_aiop_size(self):
        with self.assertRaises(ValueError):
            __ = AsyncIOPool(0)

    # async def test_aiop_cancel(self):
    #     n = 4
    #     aiop = AsyncIOPool(n)
    #     tasks = [AwaitableCounter(10) for __ in range(10)]
    #     async for __ in aiop.arun_many(tasks):
    #         aiop.cancel_all()
    #     self.assertEqual(AwaitableCounter.number_running, 0, 'no more running tasks')
    #     self.assertLess(AwaitableCounter.completed, 4, 'at most 4 completed')



from typing import Union, Awaitable, Iterable, Generator, Any, Optional
from asyncio import wait, FIRST_COMPLETED, Queue, QueueEmpty


class AsyncIOPool:
    def __init__(self, pool_size: int):
        """
        AsyncIOPool manages a queue of awaitables, starting
        :param pool_size:
        """
        self._size = 1
        # assign through setter
        self.size = pool_size

        self._tasks = Queue()
        self._aws = set()

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value: int):
        if value < 1:
            raise ValueError('AsyncIOPool.size needs to be at least 1')
        self._size = value

    async def enqueue(self, aws: Union[Awaitable, Iterable[Awaitable]]):
        """
        Add one or more awaitables to the queue of awaitable run by arun_many; more can be added while it is running
        :param aws: a single awaitable, or an iterable of awaitables to run
        :return: None
        """
        if not isinstance(aws, Iterable):
            aws = [aws]
        for aw in aws:
            await self._tasks.put(aw)

    async def arun_many(self, aws: Optional[Union[Awaitable, Iterable[Awaitable]]] = None) \
            -> Generator[Any, None, None]:
        """
        Run as many tasks as size allows in parallel, starting new ones when previous ones complete
        :param aws: a single awaitable, or an iterable of awaitables to run
        :return: a generator that yields result() from tasks as they complete
        """
        if aws is not None:
            await self.enqueue(aws)
        self._aws = set()
        while True:
            # room for more tasks and tasks queued
            while len(self._aws) < self._size and not self._tasks.empty():
                # add the next task
                self._aws.add(await self._tasks.get())
            else:
                if self._aws:
                    # run the current pool of tasks until one or more complete
                    done, self._aws = await wait(self._aws, return_when=FIRST_COMPLETED)
                    for task in done:
                        yield task.result()
            # no more awaitables, then done
            if not self._aws and self._tasks.empty():
                break

    def cancel_all(self):
        while not self._tasks.empty():
            try:
                task = self._tasks.get_nowait()
            except QueueEmpty:
                break
            task.close()
        for a in self._aws:
            a.cancel()



import unittest
from py7za import nice_size, size_to_int
from py7za._nice_size import SI_UNITS, BIN_UNITS


class TestWalker(unittest.TestCase):
    def test_nice_size_basic(self):
        self.assertEqual(nice_size(999, si=True), '999 B')
        self.assertEqual(nice_size(1000, si=True), '1.0 kB')
        self.assertEqual(nice_size(1023), '1023 B')
        self.assertEqual(nice_size(1024), '1.0 KiB')

    def test_nice_size_negative(self):
        self.assertEqual(nice_size(0), '0 B')
        self.assertEqual(nice_size(-1), '-1 B')
        self.assertEqual(nice_size(-999, si=True), '-999 B')
        self.assertEqual(nice_size(-1000, si=True), '-1.0 kB')
        self.assertEqual(nice_size(-1023), '-1023 B')
        self.assertEqual(nice_size(-1024), '-1.0 KiB')

    def test_nice_size_SI(self):
        for n, u in enumerate(SI_UNITS):
            self.assertEqual(nice_size(1000**(n+1), si=True), f'1.0 {u}')

    def test_nice_size_BIN(self):
        for n, u in enumerate(BIN_UNITS):
            self.assertEqual(nice_size(1024**(n+1)), f'1.0 {u}')

    def test_nice_size_roundtrip_bin(self):
        cases = [0, 990, 2048, 2**10, 999975760691]
        for c in cases:
            self.assertEqual(c, size_to_int(nice_size(c)))

    def test_nice_size_roundtrip_si(self):
        cases = [0, 990, 2000, 10**10]
        for c in cases:
            self.assertEqual(c, size_to_int(nice_size(c, si=True), si=True))



import unittest
from pathlib import Path
from shutil import rmtree
from py7za.py7za_box import box, CLI_DEFAULTS
from conffu import Config
from zipfile import ZipFile
from os import chdir, getcwd

CLI_DEFAULTS['output'] = 'q'


class TestPy7zaBox(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        # ensure working dir is directory of test script
        chdir(Path(__file__).parent)
        test = Path('.')
        rmtree(test / 'data')
        (test / 'data/source/sub/subsub').mkdir(parents=True, exist_ok=True)
        (test / 'data/source/sub2/').mkdir(parents=True, exist_ok=True)
        (test / 'data/source/sub3/').mkdir(parents=True, exist_ok=True)
        (test / 'data/target').mkdir(parents=True, exist_ok=True)
        (test / 'data/extra').mkdir(parents=True, exist_ok=True)
        with open('data/.gitignore', 'w') as f:
            f.write('*\n!.gitignore')
        with open('data/source/x.csv', 'w') as f:
            f.write('A,B,C\n1,2,3\n')
        with open('data/source/sub/test.txt', 'w') as f:
            f.write('Testing, 1, 2, 3')
        with open('data/source/sub/y.csv', 'w') as f:
            f.write('X,Y,Z\n0,0,0\n')
        with open('data/source/sub2/y2.csv', 'w') as f:
            f.write('X,Y,Z\n0,0,0\n')
        with open('data/source/sub3/y3.csv', 'w') as f:
            f.write('X,Y,Z\n0,0,0\n')
        # file easily mistaken for a directory
        with open('data/source/sub4', 'w') as f:
            f.write('X,Y,Z\n0,0,0\n')
        with open('data/source/sub/subsub/z.csv', 'w') as f:
            f.write('0,0,0\n')
        with open('data/source/sub/hello.txt', 'w') as f:
            f.write('Hello\nWorld\n')

    def tearDown(self) -> None:
        test = Path('.')
        rmtree(test / 'data')
        (test / 'data').mkdir(parents=True, exist_ok=True)
        with open('data/.gitignore', 'w') as f:
            f.write('*\n!.gitignore')

    async def test_box_inplace(self):
        with open('data/source/x.csv', 'rb') as f:
            original_content = f.read()
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': '*.csv', '7za': '-tzip'}))
        with ZipFile('data/source/x.csv.zip') as zf:
            with zf.open('x.csv') as f:
                self.assertEqual(original_content, f.read(), 'zipped content is identical')
        self.assertFalse(Path('data/source/x.csv').is_file(), 'original was removed')

    async def test_box_inplace_wd(self):
        wd = getcwd()
        chdir('data/source')
        with open('x.csv', 'rb') as f:
            original_content = f.read()
        await box(Config(CLI_DEFAULTS | {'glob': '*.csv', '7za': '-tzip'}))
        with ZipFile('x.csv.zip') as zf:
            with zf.open('x.csv') as f:
                self.assertEqual(original_content, f.read(), 'zipped content is identical')
        self.assertFalse(Path('x.csv').is_file(), 'original was removed')
        chdir(wd)

    async def test_box_no_delete(self):
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': '*.csv', 'delete': False}))
        self.assertTrue(Path('data/source/x.csv').is_file(), 'original was not removed')

    async def test_box_create_dirs(self):
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'target': 'data/target', 'glob': '**/*.csv',
                                         'delete': False, '7za': '-tzip'}))
        with ZipFile('data/target/x.csv.zip') as zf:
            with zf.open('x.csv') as fz:
                with open('data/source/x.csv', 'rb') as f:
                    self.assertEqual(f.read(), fz.read(), 'zipped content in root is identical')
        self.assertTrue(Path('data/target/sub/y.csv.zip').is_file(), 'file in dir zipped to subdir')
        with ZipFile('data/target/sub/y.csv.zip') as zf:
            with zf.open('y.csv') as fz:
                with open('data/source/sub/y.csv', 'rb') as f:
                    self.assertEqual(f.read(), fz.read(), 'zipped content in subdir is identical')

    async def test_box_zip_structure(self):
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'target': 'data/target', 'glob': '**/*.csv',
                                         'delete': False, 'create_dirs': False, 'zip_structure': True,
                                         '7za': '-tzip'}))
        with ZipFile('data/target/x.csv.zip') as zf:
            with zf.open('x.csv') as fz:
                with open('data/source/x.csv', 'rb') as f:
                    self.assertEqual(f.read(), fz.read(), 'zipped content in root is identical')
        self.assertTrue(Path('data/target/y.csv.zip').is_file(), 'file in dir zipped to root')
        with ZipFile('data/target/y.csv.zip') as zf:
            with zf.open('sub/y.csv') as fz:
                with open('data/source/sub/y.csv', 'rb') as f:
                    self.assertEqual(f.read(), fz.read(), 'zipped content in subdir is identical')

    async def test_box_round_trip_zip(self):
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': '*.csv', '7za': '-tzip'}))
        self.assertFalse(Path('data/source/x.csv').is_file(), 'original is gone after box')
        self.assertTrue(Path('data/source/x.csv.zip').is_file(), 'archive exists after box')
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': '*.csv.zip', 'unbox': True}))
        self.assertTrue(Path('data/source/x.csv').is_file(), 'original is back after unbox')
        self.assertFalse(Path('data/source/x.csv.zip').is_file(), 'archive removed after unbox')

    async def test_box_round_trip_7z(self):
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': '*.csv'}))
        self.assertFalse(Path('data/source/x.csv').is_file(), 'original is gone after box')
        self.assertTrue(Path('data/source/x.csv.7z').is_file(), 'archive exists after box')
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': '*.csv.7z', 'unbox': True}))
        self.assertTrue(Path('data/source/x.csv').is_file(), 'original is back after unbox')
        self.assertFalse(Path('data/source/x.csv.7z').is_file(), 'archive removed after unbox')

    async def test_box_overwrite_zip(self):
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': 'x.csv', 'delete': False, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/x.csv.zip').is_file(), 'archive exists after first box')
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': 'x.csv'}))
        self.assertFalse(Path('data/source/x.csv').is_file(), 'original is gone after box')
        with ZipFile('data/source/x.csv.zip') as zf:
            self.assertEqual(1, len(zf.filelist), 'Only one file in resulting archive')

    async def _do_test_overwrite(self, mode=None, suffix='zip'):
        with open('data/source/x.csv') as f:
            content = f.read()
        await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': 'x.csv', 'delete': False,
                                         '7za': f'-t{suffix}'}))
        with open('data/source/x.csv', 'a') as f:
            f.write('extra')
            new_content = content + 'extra'
        if mode is None:
            await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': f'x.csv.{suffix}', 'unbox': True}))
        else:
            await box(Config(CLI_DEFAULTS | {'root': 'data/source', 'glob': f'x.csv.{suffix}', 'unbox': True,
                                             'overwrite': mode}))
        return content, new_content

    async def test_box_overwrite_source_skip_default(self):
        content, new_content = await self._do_test_overwrite(suffix='zip')
        self.assertFalse(Path('data/source/x.csv.zip').is_file(), 'archive is gone after unbox')
        self.assertTrue(Path('data/source/x.csv').is_file(), '"original" file exist')
        with open('data/source/x.csv') as f:
            self.assertEqual(new_content, f.read(), 'Content of original file untouched, no overwrite')

    async def test_box_overwrite_source_all(self):
        content, new_content = await self._do_test_overwrite('a', suffix='zip')
        self.assertFalse(Path('data/source/x.csv.zip').is_file(), 'archive is gone after unbox')
        self.assertTrue(Path('data/source/x.csv').is_file(), '"original" file exist')
        with open('data/source/x.csv') as f:
            self.assertEqual(content, f.read(), 'Content of original file back to original, overwritten')

    async def test_box_overwrite_source_rename_new_zip(self):
        content, new_content = await self._do_test_overwrite('u', suffix='zip')
        self.assertFalse(Path('data/source/x.csv.zip').is_file(), 'archive is gone after unbox')
        self.assertTrue(Path('data/source/x.csv').is_file(), '"original" file exist')
        with open('data/source/x.csv') as f:
            self.assertEqual(new_content, f.read(), 'Content of original file untouched, no overwrite')
        self.assertTrue(Path('data/source/x_1.csv').is_file(), 'extracted file was renamed')
        with open('data/source/x_1.csv') as f:
            self.assertEqual(content, f.read(), 'Extracted file contains original content')

    async def test_box_overwrite_source_rename_new_7z(self):
        content, new_content = await self._do_test_overwrite('u', suffix='7z')
        self.assertFalse(Path('data/source/x.csv.zip').is_file(), 'archive is gone after unbox')
        self.assertTrue(Path('data/source/x.csv').is_file(), '"original" file exist')
        with open('data/source/x.csv') as f:
            self.assertEqual(new_content, f.read(), 'Content of original file untouched, no overwrite')
        self.assertTrue(Path('data/source/x_1.csv').is_file(), 'extracted file was renamed')
        with open('data/source/x_1.csv') as f:
            self.assertEqual(content, f.read(), 'Extracted file contains original content')

    async def test_box_overwrite_source_rename_existing_zip(self):
        content, new_content = await self._do_test_overwrite('t', suffix='zip')
        self.assertFalse(Path('data/source/x.csv.zip').is_file(), 'archive is gone after unbox')
        self.assertTrue(Path('data/source/x.csv').is_file(), '"original" file exist')
        with open('data/source/x.csv') as f:
            self.assertEqual(content, f.read(), 'extracted file to original name, original content')
        self.assertTrue(Path('data/source/x_1.csv').is_file(), 'existing file was renamed')
        with open('data/source/x_1.csv') as f:
            self.assertEqual(new_content, f.read(), 'Existing files still contains new content')

    async def test_box_overwrite_source_rename_existing_7z(self):
        content, new_content = await self._do_test_overwrite('t', suffix='7z')
        self.assertFalse(Path('data/source/x.csv.zip').is_file(), 'archive is gone after unbox')
        self.assertTrue(Path('data/source/x.csv').is_file(), '"original" file exist')
        with open('data/source/x.csv') as f:
            self.assertEqual(content, f.read(), 'extracted file to original name, original content')
        self.assertTrue(Path('data/source/x_1.csv').is_file(), 'existing file was renamed')
        with open('data/source/x_1.csv') as f:
            self.assertEqual(new_content, f.read(), 'Existing files still contains new content')

    async def test_box_dir_default(self):
        await box(Config(CLI_DEFAULTS | {'glob': '**/*.csv', '7za': '-tzip'}))
        self.assertTrue(Path('data/source/x.csv.zip').is_file(), 'archive in-place default')
        self.assertTrue(Path('data/source/sub/y.csv.zip').is_file(), 'archive subs in-place default')

    async def test_box_dir_root(self):
        await box(Config(CLI_DEFAULTS | {'root': 'data/source/sub', 'glob': '**/*.csv', '7za': '-tzip'}))
        self.assertFalse(Path('data/source/x.csv.zip').is_file(), 'no archive in root parent')
        self.assertTrue(Path('data/source/sub/y.csv.zip').is_file(), 'archive subs in-place from root')

    async def test_box_dir_target(self):
        await box(Config(CLI_DEFAULTS | {'glob': '**/*.csv', 'target': 'data/target', '7za': '-tzip'}))
        self.assertTrue(Path('data/target/data/source/x.csv.zip').is_file(),
                        'archive in target, relative to working')
        self.assertTrue(Path('data/target/data/source/sub/y.csv.zip').is_file(),
                        'archive subs in target, relative to working')
        self.assertTrue(Path('data/target/data/source/sub/subsub/z.csv.zip').is_file(),
                        'archive sub-subs in target, relative to working')

    async def test_box_dir_target_root(self):
        await box(Config(CLI_DEFAULTS | {'glob': '**/*.csv', 'target': 'data/target', 'root': 'data/source',
                                         '7za': '-tzip'}))
        self.assertTrue(Path('data/target/x.csv.zip').is_file(),
                        'archive in target, relative to root')
        self.assertTrue(Path('data/target/sub/y.csv.zip').is_file(),
                        'archive subs in target, relative to root')
        self.assertTrue(Path('data/target/sub/subsub/z.csv.zip').is_file(),
                        'archive sub-subs in target, relative to root')

    async def test_box_dirs_no_match(self):
        await box(Config(CLI_DEFAULTS | {'glob': 'sub?', 'root': 'data/source', '7za': '-tzip'}))
        self.assertFalse(Path('data/source/sub2.zip').is_file(),
                        'no zipped subdirectory without --match_dir')

    async def test_box_dirs_match(self):
        await box(Config(CLI_DEFAULTS | {'glob': 'sub?', 'root': 'data/source', 'match_dir': True, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/sub2.zip').is_file(),
                        'zipped subdirectory in-place')
        self.assertFalse(Path('data/source/sub2').is_dir(),
                         'zipped subdirectory original removed')
        self.assertTrue(Path('data/source/sub3.zip').is_file(),
                        '2nd zipped subdirectory in-place')
        await box(Config(CLI_DEFAULTS | {'glob': 'sub?.zip', 'root': 'data/source', 'unbox': True,
                                         'unbox_multi': True}))
        self.assertTrue(Path('data/source/sub2').is_dir(),
                       'zipped subdirectory restored')
        self.assertTrue(Path('data/source/sub2/y2.csv').is_file(),
                       'zipped subdirectory contents restored')
        self.assertFalse(Path('data/source/sub2.zip').is_file(),
                         'restored sub directory archive removed')
        self.assertTrue(Path('data/source/sub3').is_dir(),
                       'zipped subdirectory restored')

    async def test_box_dirs_subdirs(self):
        with open('data/source/sub/subsub/z.csv', 'rb') as f:
            content = f.read()
        await box(Config(CLI_DEFAULTS | {'glob': 'sub', 'root': 'data/source', 'match_dir': True, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/sub.zip').is_file(),
                        'archive exists')
        with ZipFile('data/source/sub.zip') as zf:
            with zf.open('sub/subsub/z.csv', 'r') as f:
                self.assertEqual(content, f.read(), 'zipped content contains files in subdirs')

    async def test_box_dirs_match_files(self):
        await box(Config(CLI_DEFAULTS | {'glob': '**/sub?', 'match_dir': True, 'match_file': True, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/sub3.zip').is_file(),
                        'zipped subdirectories in-place')
        self.assertTrue(Path('data/source/sub4.zip').is_file(),
                        'zipped files in-place')
        await box(Config(CLI_DEFAULTS | {'glob': 'sub?.zip', 'root': 'data/source', 'unbox': True,
                                         'unbox_multi': True, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/sub3').is_dir(),
                       'zipped subdirectory restored')
        self.assertTrue(Path('data/source/sub3/y3.csv').is_file(),
                       'zipped subdirectory contents restored')
        self.assertTrue(Path('data/source/sub4').is_file(),
                       'zipped files restored')
        self.setUp()
        await box(Config(CLI_DEFAULTS | {'glob': '**/sub?', 'match_dir': True, 'match_file': False, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/sub3.zip').is_file(),
                        'zipped subdirectories in-place')
        self.assertFalse(Path('data/source/sub4.zip').is_file(),
                         'no files matched and zipped')

    async def test_box_multi_glob(self):
        await box(Config(CLI_DEFAULTS) | {'glob': ['**/*.csv', '**/*.txt'], 'root': 'data', '7za': '-tzip'})
        self.assertTrue(Path('data/source/sub/test.txt.zip').is_file(),
                       'txt files matched in multi-glob')
        self.assertTrue(Path('data/source/sub/y.csv.zip').is_file(),
                       'csv files matched in multi-glob')

    async def test_zip_archives(self):
        await box(Config(CLI_DEFAULTS) | {'glob': '**/*', 'root': 'data', '7za': '-tzip'})
        self.assertTrue(Path('data/source/sub/test.txt.zip').is_file(),
                       'txt files matched in catch all')
        self.assertTrue(Path('data/source/sub/y.csv.zip').is_file(),
                       'csv files matched in catch all')
        with open('data/source/new.txt', 'w') as f:
            f.write('test')
        await box(Config(CLI_DEFAULTS) | {'glob': '**/*', 'root': 'data', '7za': '-tzip'})
        self.assertTrue(Path('data/source/new.txt.zip').is_file(),
                       'new txt files matched in catch all')
        await box(Config(CLI_DEFAULTS) | {'glob': '**/*.zip', 'unbox': True, 'root': 'data', '7za': '-tzip'})
        self.assertFalse(Path('data/source/sub/test.txt.zip').is_file(),
                         'zip files were not rezipped and')
        self.assertTrue(Path('data/source/sub/test.txt').is_file(),
                       'original content from previously created archive extracted')

    async def test_unbox_multi(self):
        await box(Config(CLI_DEFAULTS | {'glob': '**/sub', 'match_dir': True, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/sub.zip').is_file(),
                        'zipped files in sub in single archive')
        await box(Config(CLI_DEFAULTS | {'glob': '**/*.zip', 'unbox': True, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/sub.zip').is_file(),
                        'default unbox_multi False, archive not extracted')
        await box(Config(CLI_DEFAULTS | {'glob': '**/*.zip', 'unbox': True, 'unbox_multi': True, '7za': '-tzip'}))
        self.assertTrue(Path('data/source/sub/y.csv').is_file() and Path('data/source/sub/test.txt').is_file(),
                        'unbox_multi True, files extracted')
        self.assertFalse(Path('data/source/sub.zip').is_file(),
                         'unbox_multi True, archive removed')



import os
import re
from setuptools import setup

__name__ = 'py7za'

version_fn = os.path.join(__name__, "_version.py")
__version__ = "unknown"
try:
    version_line = open(version_fn, "rt").read()
except EnvironmentError:
    pass  # no version file
else:
    version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    m = re.search(version_regex, version_line, re.M)
    if m:
        __version__ = m.group(1)
    else:
        print(f'unable to find version in {version_fn}')
        raise RuntimeError(f'If {version_fn} exists, it is required to be well-formed')

with open("README.md", "r") as rm:
    long_description = rm.read()

setup(
    name="py7za",
    packages=['py7za'],
    version=__version__,
    license='MIT',
    author="BMT, Jaap van der Velde",
    author_email="jaap.vandervelde@bmtglobal.com",
    description="Description",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://gitlab.com/Jaap.vanderVelde/py7za',
    download_url='https://gitlab.com/Jaap.vanderVelde/py7za/repository/archive.zip?ref='+__version__,
    keywords=['package'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
    extras_require={
        'box': ['conffu'],
        'dev': ['mkdocs']
    },
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'py7za-box=py7za.py7za_box:cli_entry_point',
            'box=py7za.py7za_box:cli_box_entry_point',
            'unbox=py7za.py7za_box:cli_unbox_entry_point',
        ],
    },
    include_package_data=True,
    package_data={'': ['bin/7za.exe', 'bin/7za-license.txt', 'bin/GNU_LGPL.txt']},
)



from shlex import split
from sys import stdout
from os import remove as os_remove, stat, name as os_name
from datetime import datetime, timezone
import logging
from logging import error, warning, info
from conffu import Config
from pathlib import Path
from asyncio import get_event_loop, sleep, gather
from py7za import Py7za, AsyncIOPool, available_cpu_count, nice_size
from subprocess import list2cmdline
import re
from json import load
from json.decoder import JSONDecodeError
from typing import Optional
# only used for status
from zipfile import ZipFile

ARCHIVE_SUFFIXES = ['.zip', '.zipx', '.gz', '.7z', '.s7z', '.lzma', '.lz', '.cab', '.jar', '.war', '.rar']

aiop: Optional[AsyncIOPool] = None


async def box(cfg):
    global aiop

    start_time = datetime.now()
    total = 0
    done = False
    skipped = 0
    finished = 0
    total_content_size = 0
    total_zip_size = 0
    current = 0
    running = []

    if 'regex' in cfg:
        try:
            regex = re.compile(cfg.regex)
        except re.error:
            error(f'Invalid regex pattern: "{cfg.regex}"')
            exit(1)
    else:
        regex = False

    parallel = 0
    try:
        if isinstance(cfg['parallel'], str) and cfg['parallel'] and cfg['parallel'][-1].lower() == 'x':
            parallel = int(float(cfg['parallel'][:-1]) * available_cpu_count())
        else:
            parallel = int(cfg['parallel'])
    except ValueError:
        error(f'Invalid value for --parallel: {cfg["parallel"]}')
        exit(1)
    if parallel == 0:
        parallel = available_cpu_count()
    elif parallel < 0:
        parallel = available_cpu_count() + parallel
    if parallel < 1:
        parallel = 1
    aiop = AsyncIOPool(pool_size=parallel)

    cli_options = '-y '
    cli_options += '-sdel ' if cfg['delete'] and not cfg['unbox'] else ''
    cli_options += f'-ao{cfg.overwrite} ' if cfg['unbox'] else ''
    cli_options += cfg['7za'] if '7za' in cfg else ''
    types = [o for o in split(cli_options) if o.startswith('-t')]
    if len(types) > 1:
        error(f'Cannot provide multiple archive types to 7za: {types}')
        exit(1)
    elif len(types) == 0:
        cfg.suffix = '.7z'
    else:
        cfg.suffix = f'.{types[0][2:]}'

    print_result = cfg.output in 'dlv'
    info_command = cfg.output == 'v'
    level = logging.INFO if info_command else logging.WARNING
    if 'log_error' in cfg:
        # only log to file
        logging.basicConfig(
            level=level, format='%(asctime)s %(levelname)s %(message)s', filename=cfg.log_error, filemode='a')
    else:
        # only log to console, no additional setup required
        logging.basicConfig(
            level=level, format='%(asctime)s %(levelname)s %(message)s')
    update_status = cfg.output == 's'
    list_status = cfg.output == 'd'

    log = cfg['log'] if 'log' in cfg else None
    test = cfg['test'] if 'test' in cfg else False
    skipped_test = 0

    if cfg['match_groups']:
        groups_file = Path(__file__).parent / 'groups.json'
        if not groups_file.is_file():
            warning(f'Group matching specified, but no groups.json found in script installation directory.')
        else:
            cfg['group_map'] = group_map(groups_file, cfg)

    unbox = cfg['unbox']
    from_into = "from" if unbox else "into"
    unbox_multi = cfg['unbox_multi']
    delete = cfg['delete']
    create_dirs = cfg['create_dirs']
    zip_structure = cfg['zip_structure']
    zip_archives = cfg['zip_archives']
    match_groups = cfg['match_groups']
    si = cfg['si']

    def globber(root, glob_expr):
        nonlocal skipped, regex
        group_results = set()
        if not isinstance(glob_expr, list):
            glob_expr = [glob_expr]
        for ge in glob_expr:
            for fn in Path(root).glob(ge):
                if regex and not regex.match(str(fn)):
                    info(f'Skipping {fn} as it does not match provided regex.')
                    skipped += 1
                    continue
                if (fn.is_file() and cfg['match_file']):
                    if not unbox and not zip_archives and fn.suffix in ARCHIVE_SUFFIXES:
                        info(f'Skipping {fn} as it is an archive and --zip_archives was not specified.')
                        skipped += 1
                        continue
                    if (fn.relative_to(root).parent, fn.name) not in group_results:
                        if match_groups and (
                            (not unbox and fn.suffix in cfg['group_map']) or
                            (unbox and Path(fn.stem).suffix in cfg['group_map'])
                        ):
                            parent, stem, suffix, extra = (fn.parent, fn.stem, fn.suffix, '') \
                                if not unbox else (fn.parent, Path(fn.stem).stem, Path(fn.stem).suffix, fn.suffix)
                            matches = []
                            for suffixes in cfg['group_map'][suffix]:
                                matches = [
                                    parent / f'{stem}{sf}{extra}'
                                    for sf in suffixes
                                    if (parent / f'{stem}{sf}{extra}').is_file()
                                ]
                                if matches:
                                    break
                            for group_fn in matches:
                                parent, name = group_fn.relative_to(root).parent, group_fn.name
                                if (parent, name) not in group_results:
                                    group_results.add((parent, name))
                                    yield parent, name
                        yield fn.relative_to(root).parent, fn.name
                elif (fn.is_dir() and cfg['match_dir']):
                    yield fn.relative_to(root).parent, fn.name

    def start(py7za):
        nonlocal running, current, info_command
        if info_command:
            info(f'"{py7za.working_dir}": '+list2cmdline([py7za.executable_7za, *py7za.arguments]))
        current += 1
        running.append(py7za)

    async def print_status():
        nonlocal done, running, current, total, from_into
        while True:
            if total == 0:
                stdout.write(f'\x1b[2K\rStarting ... ')
            else:
                stdout.write(f'\x1b[2K\rProcessing: {current} / {total} '
                             f'[{nice_size(total_content_size, si)} {from_into} {nice_size(total_zip_size, si)}] ... '
                             f'{" ".join([f"{str(py7za.progress)}/100%" for py7za in running if py7za.progress])}')
            stdout.flush()
            if done:
                return
            await sleep(0.5)

    async def run_all():
        nonlocal total, done, total_content_size, total_zip_size, skipped, finished, test, skipped_test
        global aiop

        zippers = []

        root = Path(cfg.root).absolute()
        target = Path(cfg.target).absolute() if 'target' in cfg else root
        n = 0
        try:
            for sub_path, fn in globber(cfg.root, cfg.glob):
                if print_result and n % 100 == 0 and not test:
                    stdout.write(f'\x1b[2K\rMatching [{n}] ... ')
                n += 1
                if create_dirs and not test:
                    (target / sub_path).mkdir(parents=True, exist_ok=True)
                if not unbox:
                    target_path = target / sub_path / fn if create_dirs else target / fn
                    if zip_structure:
                        content = sub_path / fn
                        wd = str(root)
                    else:
                        content = root / sub_path / fn
                        wd = '.'
                    if not test:
                        zippers.append(
                            Py7za(f'a "{target_path}{cfg.suffix}" "{content}" {cli_options}',
                                  on_start=start, working_dir=wd))
                    else:
                        skipped_test += 1
                        print(f'TEST in "{wd}": 7za.exe a "{target_path}{cfg.suffix}" "{content}" {cli_options}')
                else:
                    archive = root / sub_path / fn
                    if not unbox_multi and (
                            (archive.suffix == '.zip' and len(ZipFile(archive).filelist) > 1) or
                            ((await Py7za.list_archive(str(archive), meta_data_only=True))[2] > 1)):
                        info(f'Skipping {archive} as it has multiple files and --unbox_multi was not specified.')
                        skipped += 1
                        continue
                    target_path = target / sub_path if create_dirs else target
                    if not test:
                        zippers.append(Py7za(f'x "{archive}" "-o{target_path}" {cli_options}', start))
                    else:
                        skipped_test += 1
                        print(f'TEST: 7za.exe x "{archive}" "-o{target_path}" {cli_options}')
            if print_result:
                stdout.write(f'\x1b[2K\rMatched {n} object(s), '
                             f'start processing in up to {aiop.size} parallel processes ...\n')
            total = len(zippers)
            async for py7za in aiop.arun_many(zippers):
                finished += 1
                if py7za.return_code > 0:
                    error(f'Return code {py7za.return_code}'
                          f' from: {list2cmdline([py7za.executable_7za, *py7za.arguments])}'
                          f'\n{py7za.errors.decode()}')
                    continue
                fn = py7za.arguments[1]
                pfn = Path(fn)
                if not pfn.is_file():
                    error(f'Archive file {fn} not found.')
                    continue
                if print_result or update_status or log:
                    if pfn.suffix == '.zip':
                        total_content_size += (cs := sum([zip_info.file_size for zip_info in ZipFile(fn).filelist]))
                    else:
                        total_content_size += (cs := (await Py7za.list_archive(fn, meta_data_only=True))[0])
                    total_zip_size += (zs := stat(fn).st_size)
                    if log:
                        with open(log, 'a') as lf:
                            lf.write(
                                f'{datetime.strftime(datetime.now(timezone.utc).astimezone(), "%Y-%m-%d %H:%M:%S%z")},'
                                f'{fn},{nice_size(cs, si)},{cs},{nice_size(zs, si)},{zs}\n')
                    if print_result:
                        if list_status:
                            stdout.write(f'\x1b[2K\r'
                                         f'{datetime.strftime(datetime.now(), "%H:%M:%S")}  '
                                         f'{nice_size(cs, si)} {from_into} {nice_size(zs, si)} {fn}\n')
                            stdout.write(f'Total: {finished} / {total} '
                                         f'[{nice_size(total_content_size, si)} '
                                         f'{from_into} {nice_size(total_zip_size, si)}] ({current - finished} running)')
                        else:
                            stdout.write(f'{datetime.strftime(datetime.now(), "%H:%M:%S")}  '
                                         f'{nice_size(cs, si)} {from_into} {nice_size(zs, si)} {fn}\n')
                        stdout.flush()
                if unbox and delete:
                    os_remove(py7za.arguments[1])
                running.remove(py7za)
            done = True
        except (KeyboardInterrupt, RuntimeError):
            pass

    if update_status:
        await gather(run_all(), print_status())
    else:
        await run_all()

    if cfg.output != 'q':
        if update_status or print_result:
            stdout.write('\x1b[2K\r')
        print(f'Completed processing {nice_size(total_content_size)} '
              f'of files {"from" if unbox else "into"} {total} archives, totaling {nice_size(total_zip_size)}.'
              f'\nTook {datetime.now()-start_time}. Skipped {skipped} files matching glob.')
        if test:
            print(f'TEST: would have started {"un" if unbox else ""}boxing {skipped_test} matches.')


def print_help():
    from ._version import __version__
    print(
        '\npy7za-box '+__version__+', command line utility\n'
        '\nPy7za-box ("pizza box") replaces a set of files with individual .zip files\n'
        'containing the originals, or does the reverse by "unboxing" the archives.\n'
        'Py7za uses 7za.exe, more information on the project page.\n'
        '\n'
        'Usage: `py7za-box | box <glob pattern(s)> [options]`\n'
        '\n'
        '<glob pattern(s)>         : Glob pattern(s) like "**/*.csv". (required)\n'
        '                            Add quotes if your pattern contains spaces.\n'
        'Options:\n'
        '-h/--help                 : This text.\n'
        '-d/--delete               : Remove the source after (un)boxing. [True]\n'
        '-cd/--create_dirs         : Recreate dir structure in target path. [True]\n'
        '-l/--log <path>           : Log source, size, target, size as .csv. [None]\n'
        '-el/--error_log <path>    : Log warnings and error messages to file. [None]\n'
        '-gm/--group_match [bool]  : Group files with grouped suffixes. [True]\n'
        '-ga/--group_add <path>    : Path to extra .json group definitions. [None]\n'
        '-md/--match_dir [bool]    : Glob pattern(s) should match dirs. [False]\n'
        '-mf/--match_file [bool]   : Glob pattern(s) should match files. [True]\n'
        '-p/--parallel <n>         : #Parallel processes to run [0 = available cores]\n'
        '-r/--root <path>          : Path glob pattern(s) are relative to. ["."]\n'
        '-re/--regex <expr>        : Regex path of globbed files must match. [None]\n'
        '-t/--target <path>        : Root path for output. ["" / in-place]\n'
        '-u/--unbox/--unzip        : Unzip instead of zip (glob to match archives).\n'
        '-um/--unbox_multi         : Whether to unzip multi-file archives. [False]\n'
        '                            (implies --unbox, which can be omitted)\n'
        '-o/--output [d/l/q/s/v]   : Default (a line per archive with status), list,\n'
        '                            quiet, status, or verbose output. Verbose prints\n'
        '                            each full 7za command and logs at info level.\n'
        '                            Note: d/l/v do not work for all archive types.\n'
        '-si                       : Whether to use SI units for file sizes. [False]\n'
        '--test                    : Test mode - no files are changed. [False]\n'
        '-w/--overwrite [a/s/u/t]  : Used overwrite mode when unboxing. [s]\n'
        '                            a:all, s:skip, u:rename new, t:rename existing.\n'
        '-za/--zip_archives [bool] : Whether to zip matched archives (again). [False]\n'
        '-zs/--zip_structure [bool]: Root subdirectory structure is archived. [False]\n'
        '-7/--7za <arguments>      : CLI arguments passed to 7za after scripted ones.\n'
        '                            Add quotes if passing more than one argument.\n'
        '\n'
        'Examples:\n'
        '\n'
        'Zip all .csv files in C:/Data and put the archives in C:/Archive:\n'
        '   py7za-box *.csv --root C:/Data --target C:/Archive\n'
        'Unzip all .csv.zip from C:/Archive and subdirectories to C:/Data:\n'
        '   py7za-box **/*.csv.zip --unbox --root C:/Archive -t C:/Data\n'
        '   unbox **/*.csv.zip --root C:/Archive -t C:/Data\n'
        'Zip directories named `Photo*` individually using maximum compression:\n'
        '   box Photo* -r "C:/My Photos" -md -mf 0 -t C:/Archive -7 "-mx9"\n'
        '\nNote that you can gracefully interrupt a (un)boxing run with Ctrl+C.\n'
    )


CLI_DEFAULTS = {
    'parallel': '0',
    'delete': True,
    'create_dirs': True,
    'match_groups': True,
    'match_dir': False,
    'match_file': True,
    'output': 'd',
    'overwrite': 's',
    'si': False,
    'root': '.',
    'test': False,
    'unbox': False,
    'unbox_multi': False,
    'verbose': False,
    'zip_structure': False,
    'zip_archives': False,
    '7za': ''
}


def group_map(fn, cfg):
    try:
        with open(str(fn), 'r') as f:
            groups = load(f)
    except (IOError, JSONDecodeError) as e:
        error(f'Error reading groups.json for group matching: {e}')
    if 'group_add' in cfg:
        if not Path(cfg['group_add'].is_file()):
            warning(f'Additional group definitions specified, file was not found: {cfg["group_add"]}')
        try:
            with open(str(cfg['group_add']), 'r') as f:
                added_groups = load(f)
            groups = groups | added_groups
        except (IOError, JSONDecodeError) as e:
            error(f'Error reading {cfg["group_add"]} additional group definitions: {e}')
    elif 'groups' in cfg:
        groups = groups | cfg['groups']
    return {
        suffix: [
            [other for other in g if other != suffix] for g in groups.values() if suffix in g
        ] for suffix in [s for v in groups.values() for s in v]
    }


def cli_entry_point(unbox=False):
    global aiop

    # enable sufficient ansi support on Windows
    if os_name == 'nt':
        from ctypes import windll
        k = windll.kernel32
        k.SetConsoleMode(k.GetStdHandle(-11), 7)

    cfg = Config.startup(defaults=CLI_DEFAULTS, no_compound_keys=True, aliases={
        'h': 'help', 'p': 'parallel', 'cd': 'create_dirs', 'md': 'match_dir', 'mf': 'match_file', 'u': 'unbox',
        'unzip': 'unbox', 'r': 'root', 'zs': 'zip_structure', 't': 'target', 'v': 'verbose', '7': '7za', 'g': 'glob',
        'o': 'output', 'w': 'overwrite', 'za': 'zip_archives', 'um': 'unbox_multi', 'l': 'log', 'le': 'log_error',
        'unzip_multi': 'unbox_multi', 'error_log': 'log_error', 'el': 'log_error', 're': 'regex',
        'regular_expression': 'regex', 'mg': 'match_groups', 'ga': 'group_add', 'cf': 'create_dirs',
        'create_folders': 'create_dirs'
    })

    if unbox:
        if 'unbox' in cfg.arguments and not cfg.unbox:
            warning('"--unbox False" option passed to `unbox` ignored, unboxing.')
        cfg.unbox = True

    if cfg.get_as_type('help', bool, False):
        print_help()
        exit(0)

    if len(cfg.arguments['']) < 2 and 'glob' not in cfg:
        error('Missing required argument glob pattern(s), e.g. `py7za-box *.csv [options]`.')
        print_help()
        exit(1)
    else:
        cfg.glob = cfg.glob if 'glob' in cfg else cfg.parameters

    if not Path(cfg.root).is_dir():
        error(f'The provided root directory "{cfg.root}" was not found.')
        exit(2)

    target = cfg.root if 'target' not in cfg or cfg.target is True else cfg.get_as_type('target', str)
    if not Path(target).is_dir():
        error(f'The provided target directory "{cfg.target}" was not found.')
        exit(2)

    if cfg.unbox and cfg.create_dirs and 'target' in cfg and target is not True:
        warning(f'When unboxing to a target location (not in-place), original structure cannot be restored '
                f'unless subdirectory from root were included when the archives were created.')

    if cfg.zip_structure and cfg.unbox:
        warning(f'The --zip_structure option was specified, but does nothing when unboxing and will be ignored.')

    if cfg.zip_archives and cfg.unbox:
        warning(f'The --zip_archives option was specified, but does nothing when unboxing and will be ignored.')

    if cfg.unbox_multi and not cfg.unbox:
        if 'unbox' in cfg.arguments:
            warning(f'The --unbox_multi option was specified, but unbox was set to False, so the option is ignored.')
        else:
            # set unbox to true if only unbox_multi was provided
            cfg.unbox = True

    if cfg.zip_structure and cfg.create_dirs:
        warning(f'Keeping subdirectories from root in archives, as well creating the directory structure in the '
                f'target location may produce unexpected results.')

    output_modes = {
        'd': 'd', 'default': 'd',
        'l': 'l', 'list': 'l',
        'q': 'q', 'quiet': 'q',
        's': 's', 'status': 's',
        'v': 'v', 'verbose': 'v'
    }
    if cfg.verbose:
        if 'output' in cfg.arguments:
            warning(f'Setting --verbose to True overrides any --output options with \'verbose\'.')
        cfg.output = 'v'
    if cfg.output not in output_modes:
        error(f'Unknown output mode {cfg.output}, provide default(d), list(l), quiet(q), status(s) or verbose(v).')
        print_help()
        exit(1)
    cfg.output = output_modes[cfg.output]

    overwrite_modes = {
        'a': 'a', 'all': 'a',
        's': 's', 'skip': 's',
        'u': 'u', 'rename_new': 'u',
        't': 't', 'rename_existing': 't'
    }
    if cfg.overwrite not in overwrite_modes:
        error(f'Unknown overwrite mode {cfg.output}, provide all(a), skip(s), rename_new(u) or rename_existing(t).')
        print_help()
        exit(1)
    cfg.overwrite = overwrite_modes[cfg.overwrite]
    if cfg.overwrite != 's' and not cfg.unbox:
        warning(f'Overwrite mode {cfg.overwrite} passed, but option will have no effect unless unboxing.')

    loop = get_event_loop()
    try:
        loop.run_until_complete(box(cfg))
    except KeyboardInterrupt:
        print('\nExecution interrupted, cleaning up...')
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        # not closing the loop, to avoid further exceptions - we're on the way out now


def cli_box_entry_point():
    cli_entry_point()


def cli_unbox_entry_point():
    cli_entry_point(unbox=True)


if __name__ == '__main__':
    cli_entry_point()
